open Mllexeraux
open Mllexer


let end_of_transmission = Char.chr 255;;


(**********************************************************************)
(* LEXICOGRAPHIC ANALYSIS *********************************************)
(**********************************************************************)

let position = ref 0;;
let line = ref 0;;
let line_of_token = ref 0;;
let line_position = ref 0;;



let print_position () =
  prerr_string ("In line " ^ (string_of_int !line) ^ 
                " character " ^ (string_of_int (!position - !line_position)) ^
                ":");
  prerr_newline ();
;;


let filter_nulltokens (analyzer: Lexing.lexbuf->token) lexbuf =
  let rec r t =
  (     
    match t with
      Nulltoken    -> line_of_token := !line;
                      r (analyzer lexbuf)
    | Lines (n,t') -> line_of_token := !line;
                      line := !line + n;
                      line_position := (Lexing.lexeme_start lexbuf)+1;
                      r t'
    | _            -> position := Lexing.lexeme_start lexbuf;
                      t
  ) in r (analyzer lexbuf)
;;


let get file =
  line := 1;
  line_of_token := 1;
  line_position := 0;
  let ch = open_in file in
  let lexbuf = Lexing.from_channel ch in
  let get_token' = filter_nulltokens get_token in
  let rec schleife () =
    let t = get_token' lexbuf in
    let n = !line_of_token in
    if t = Endofstream then
      [n, t]
    else
      (n, t) :: (schleife())
  in
    let l = schleife () in
    close_in ch;
    l
;;


(**********************************************************************)


let rec write_token_list tl =
  match tl with
    []       -> ()
  | (n,x) :: tl' 
             -> print_string ("line " ^ string_of_int n ^ ": ");
                ( match x with 
                    Ident s             -> print_string ("|" ^ s ^ "|\n")
                  | Integer_constant i  -> print_int i; print_string "\n"
                  | Tok_constant s      -> print_string ("`" ^ s ^ "`\n")
                  | Var_constant s      -> print_string ("var " ^ s ^ "\n")
                  | Toklist_constant sl -> List.map (fun y -> 
                                                  print_string("`" ^ y ^ "` "))
                                               sl;
                                           print_string "\n"
                  | String_constant s   -> print_string ("\"" ^ s ^ "\"\n")
                  | Comment s           -> print_string ("%" ^ s ^ "%\n")
                  | Symbol s            -> print_string ("   " ^ s ^ "\n")
                  | _                   -> ()
                );
                write_token_list tl'
;;





(**********************************************************************)
(********** EXTRACT OBJECT AND TYPE DEFINITIONS FROM PHRASES **********)
(**********************************************************************)


let rec remove_comments tl =
  match tl with
    []           -> []
  | (n,x) :: tl' -> ( match x with
                        Comment c -> remove_comments tl'
                      | _         -> (n,x) :: remove_comments tl'
                    )
;;


let split_into_toplevel_phrases tl =
  let rec split_rec phrase tl =
  ( match tl with
      []           -> []        (* phrase is simply dropped !!! *)
    | (n,x) :: tl' -> if x=Symbol ";;" then
                        (phrase @ [n,x]) :: 
                        (split_rec [] tl')
                      else
                        split_rec (phrase @ [n,x]) tl'
  ) in
  split_rec [] tl
;;



(* TODO: weitere Klammern beruecksichtigen:
 * let ... in        und Derivate
 * abstype ... in    und Derivate
 *)

(* QUESTIONABLE:
 * let a =
 *   ... x ... y ...
 *   where x = ...
 *     and y = ...;;
 * Das "and" wird immer gelesen wie: let a = (...x...y...where x=) and y =...
 * KLAEREN OB DAS KORREKT IST.
 *)

(* tl sei syntaktische Struktur 
 * <Klausel> and <Klausel> and ... and <Klausel>
 * f wird fuer alle Klauseln aufgerufen und die von f berechneten Listen
 * werden konkateniert. 
 * Klammern in Klauseln werden beachtet, z.B. wird f bei
 * tl = K1 and ... (... and K2 ...) ... and K3
 * nur fuer K1, "... (... and K2 ...) ..." und K3 aufgerufen.
 *)
let rec for_all_clauses f tl =
  let r = f tl in

  let rec find_and paren tl =
  ( match tl with
      []           -> []
    | (n,x) :: tl' -> ( match x with
                          Ident "and" -> if paren = 0 then
                                            for_all_clauses f tl'
                                         else
                                            find_and paren tl'
                        | (Symbol "(" |
                           Symbol "[" |
			   Ident "let" |
                           Ident "letref" |
                           Ident "letrec" |
                           Ident "lettype" |
			   Ident "abstype" |
                           Ident "absrectype") -> find_and (paren+1) tl'
                        | (Symbol ")" |
                           Symbol "]" |
			   Ident "in")  -> find_and (paren-1) tl'
                        | _           -> find_and paren tl'
                      )
  ) in

  r @ find_and 0 tl
;;



(* bestimmt tl2 mit tl = tl1 @ tl2 so, dass tl2 mit einem Token aus sl
 * beginnt und tl1 moeglichst kurz. Es wird Klammerung beachtet.
 *)
let find_token sl tl =
 
  let rec find_rec paren tl =
  ( match tl with
      []           -> failwith "unknown syntax"
    | (n,x) :: tl' -> if paren = 0 & List.mem x sl then
                        tl
                      else
                      ( match x with
                         (Symbol "(" |
                           Symbol "[" |
			   Ident "let" |
                           Ident "letref" |
                           Ident "letrec" |
                           Ident "lettype" |
			   Ident "abstype" |
                           Ident "absrectype") -> find_rec (paren+1) tl'
                        | (Symbol ")" |
                           Symbol "]" |
			   Ident "in")   -> find_rec (paren-1) tl'
                        | _           -> find_rec paren tl'
                      )
  ) in
  
  find_rec 0 tl
;;


(* bestimmt tl1 mit tl = tl1 @ tl2 so, dass tl2 mit einem Token aus sl
 * beginnt und tl1 moeglichst kurz. Es wird Klammerung beachtet.
 *)
let find_before_token sl tl =
  let tl' = List.rev tl in
  let tl1'_plus = find_token sl tl' in
  let tl1 = List.rev (List.tl tl1'_plus) in
    tl1
;;


(* Vorbed.: tl enthaelt keine 'Lines', 'Nulltoken', 'Endofstream' und
 * 'Comment';
 * tl ist eine Toplevel-Phrase
 * ZURUECK: (int * string) list = zeilennumerierte Bezeichner
 *)
let extract_object_definitions tl =
  let rec extract_odef_from_pattern tl =
  ( 
    let extract_odef_from_cons_pattern tl =   (* p1.p2 oder p1,p2 *)
    ( let (n,x) = List.hd tl in
      match x with
        Symbol "." -> extract_odef_from_pattern (List.tl tl)
      | Symbol "," -> extract_odef_from_pattern (List.tl tl)
      | _          -> ([], tl)
    ) in

    let rec extract_odef_from_list_pattern tl =   (* [p1;...;pn] *)
    ( let (n,x) = List.hd tl in
      match x with
        Symbol "]" -> ([], List.tl tl)
      | Symbol ";" -> extract_odef_from_list_pattern (List.tl tl)
      | _          -> let (l1, tl') = extract_odef_from_pattern tl in
                      let tl'' = find_token [Symbol ";"; Symbol "]"] tl' in
                      let (l2, tl''') = extract_odef_from_list_pattern tl'' in
                        (l1 @ l2, tl''')
    ) in

    let extract_odef_from_paren_pattern tl =      (* () oder (p) *)
    ( let (n,x) = List.hd tl in
      match x with
        Symbol ")" -> ([], List.tl tl)
      | _          -> let (l1, tl') = extract_odef_from_pattern tl in
                      let tl'' = find_token [ Symbol ")"] tl' in
                        (l1, List.tl tl'')
    )  in

    let (n,x) = List.hd tl in
    let (l1, tl') =
      ( match x with
          Symbol "(" -> extract_odef_from_paren_pattern (List.tl tl)
        | Symbol "[" -> extract_odef_from_list_pattern (List.tl tl)
        | Ident s    -> ([s] , List.tl tl)
        | _          -> ([]  , tl)
      ) in 
    let (l2, tl'') = extract_odef_from_cons_pattern tl' in
      (l1 @ l2, tl'')
  ) in


  let extract_odef_from_binding tl =
  ( for_all_clauses
      (fun tl1 -> fst (extract_odef_from_pattern tl1)
      )
      tl
  ) in


  let extract_odef_from_abs_binding tl =
  ( let tl' = List.tl (find_token [Ident "with"] tl) in
    extract_odef_from_binding tl'
  ) in


  let (n,x) = List.hd tl in
  match x with
    Ident "let"        -> extract_odef_from_binding (List.tl tl)
  | Ident "letref"     -> extract_odef_from_binding (List.tl tl)
  | Ident "letrec"     -> extract_odef_from_binding (List.tl tl)
  | Ident "abstype"    -> extract_odef_from_abs_binding (List.tl tl)
  | Ident "absrectype" -> extract_odef_from_abs_binding (List.tl tl)
  | Ident "lettype"    -> []
  | _                  -> []
;;



(* Vorbed.: tl enthaelt keine 'Lines', 'Nulltoken', 'Endofstream' und
 * 'Comment';
 * tl ist eine Toplevel-Phrase
 * ZURUECK: (int * string) list = zeilennumerierte Bezeichner
 *)
let extract_type_definitions tl =

  let extract_tdef_from_abs_binding tl =
  ( let tl_before_with = find_before_token [Ident "with"] tl in
    for_all_clauses
      (fun tl1 -> let tl2 = find_before_token [Symbol "="] tl1 in
	          if tl2 = [] then failwith "unkwnown syntax" else
                    let (n,x) = List.hd (List.rev tl2) in
                    match x with
                      Ident s -> [s]
                    | _       -> failwith "unknown syntax"
      )
      tl_before_with
  ) in

  let extract_tdef_from_lettype tl =
  ( for_all_clauses
      (fun tl1 -> if tl1 = [] then failwith "unknown syntax" else
                  match snd(List.hd tl1) with
                    Ident s -> [s]
                  | _       -> failwith "unknown syntax"
      )
      tl
  ) in

  let (n,x) = List.hd tl in
  match x with
    Ident "let"        -> []
  | Ident "letref"     -> []
  | Ident "letrec"     -> []
  | Ident "abstype"    -> extract_tdef_from_abs_binding (List.tl tl)
  | Ident "absrectype" -> extract_tdef_from_abs_binding (List.tl tl)
  | Ident "lettype"    -> extract_tdef_from_lettype (List.tl tl)
  | _                  -> []
;;


(**********************************************************************)
(**************** OBTAINING COMMENTS DESCRIBING A PHRASE **************)
(**********************************************************************)

(* Ein Kriterium, wann c ein beschreibender Kommentar ist und in der
 * Analyse ausgegeben wird.
 * Kann den Beduerfnissen angepasst werden.
 *)
let is_describing_comment c =
  (String.length c >= 2) &
  (String.sub c 0 2 = "%!")
;;


(* Gibt die Liste der beschreibenden Kommentare fuer die Phrase zurueck,
 * die durch tl angegeben ist.
 * tl soll keine Nulltoken, Lines und Endofstream enthalten.
 * Es werden nur diejenigen Kommentare zurueckgegeben, die am Beginn der
 * Phrase stehen und das Kriterium von is_describing_comment erfuellen.
 *) 
let rec get_describing_comments tl =
  match tl with
    []                   -> []
  | (n,Comment s) :: tl' -> if is_describing_comment s then
                              s :: (get_describing_comments tl')
                            else
                              get_describing_comments tl'
  | tl'                  -> []
;;


(**********************************************************************)
(**** GETTING THE SET OF ALL IDENTIFIERS (NON-KEYWORD) OF A PHRASE  ***)
(**********************************************************************)


let keywords = 
  ["let"; "letrec"; "letref"; "abstype"; "absrectype"; "lettype";
   "where"; "whererec"; "whereref"; "wheretype"; "whereabstype"; 
   "whereabsrectype"; "in";
   "and"; "or"; "do"; "fail"; "failwith"; "if"; "then"; "loop"; "else";
   "not"; 
   "int"; "unit"; "bool"; "tok"; "var"; "list"; "string"
  ];;


let keyword_table =
  let kt = Hashtbl.create 997 in
    List.iter (fun x -> Hashtbl.add kt x ()) keywords;
    kt
;;


let get_idents tl =
  let rec collect tl =
  ( match tl with
      []                   -> []
    | (n,Ident s) :: tl'   -> s :: (collect tl')
    | _ :: tl'             -> collect tl'
  ) in
  let rec make_set sl =
  ( match sl with
      []       -> []
    | s :: sl' -> let sl'_set = make_set sl' in
                  try Hashtbl.find keyword_table s; sl'_set
                  with
                    Not_found -> if List.mem s sl'_set then sl'_set
                                                       else s :: sl'_set
  ) in
  
  make_set (collect tl)
;;


(**********************************************************************)
(************ ANALYSIS OF PHRASES: COMPILING THE RESULTS **************)
(**********************************************************************)


type phrase =
{ obj_defs : (int * string) list        (* Objektdefinitionen inkl Zeilennr.*)
; type_defs : (int * string) list       (* Typdefinitionen inkl. Zeilennr. *)
; comments : string list                (* Kommentare zur Phrase *)
; ident_set : string list               (* Vorkommende Bezeichner *)
; pline : int
}
;;


let get_phrases tl =
  let pl = split_into_toplevel_phrases tl in
  let defll = List.map (fun p -> 
			  let cl = get_describing_comments p in
			  let p' = remove_comments p in
			  let (n,x) = List.hd p' in
                          let odefs, tdefs  = try 
                                        extract_object_definitions p',
					extract_type_definitions p'
                                      with
                                       Failure f -> 
				         prerr_string ("Phrase in line " ^ 
						       string_of_int n ^ 
						       ": " ^ f);
					 prerr_newline();
					 [], []
                                     in
			  { obj_defs  = List.map (fun d -> (n,d)) odefs
			  ; type_defs = List.map (fun d -> (n,d)) tdefs
			  ; comments = cl
			  ; ident_set = get_idents p'
			  ; pline = n
			  }
                       )
                       pl in
  defll
;;


(**********************************************************************)
(********************* FILE DESCRIPTIONS ******************************)
(**********************************************************************)


type filedesc =
{ fname : string
; ph : phrase list
; mutable direct_deps : int list
};;


let get_filedesc fname =
  prerr_string ("* scanning and analyzing file " ^ fname);
  prerr_newline();
  try
    let tl = get fname in
      { fname = fname
      ; ph = get_phrases tl
      ; direct_deps = []
      }
  with
    Failure "lexing: empty token" -> 
      print_position();
      prerr_string "bad token";
      prerr_newline();
        { fname = fname
        ; ph = []
        ; direct_deps = []
        }
;;


let files = (ref [| |] : filedesc array ref);;
  (* alle Files der Kommandozeile *)


let set_files fnames =
  let i = ref 0 in
  let dummy = { fname = ""; ph = []; direct_deps = [] } in
  files := Array.create (List.length fnames) dummy;
  List.iter
    ( fun x ->
        ( !files ).( !i ) <- get_filedesc x;
        incr i
    )
    fnames
;;


(**********************************************************************)
(* MAPPING IDENTIFIERS TO DEFINITIONS *********************************)
(**********************************************************************)


let ignored_idents = (ref [] : string list ref);;

type obj_or_type = Odef | Tdef;;

type namedesc =
  { kind : obj_or_type
  ; file : int             (* index in !files *)
  ; line : int
  }
;;


let nametable = (Hashtbl.create 997 : (string, namedesc) Hashtbl.t);;
(* global hash table of all identifiers *)


let clear_nametable () =
  Hashtbl.clear nametable
;;


let enter_phrase i ph =
  let odefs = ph.obj_defs in
  let tdefs = ph.type_defs in
  List.iter 
    (fun (n,x) ->
       if not (List.mem x !ignored_idents) then
         Hashtbl.add nametable x { kind = Odef; file = i; line = n }
    )
    odefs;
  List.iter 
    (fun (n,x) ->
       if not (List.mem x !ignored_idents) then
         Hashtbl.add nametable x { kind = Tdef; file = i; line = n }
    )
    tdefs
;;



(* sorted list of all identifiers in nametable *)
let sorted_list () =
  let l = ref [] in
  Hashtbl.iter 
    (fun s nd ->
       l := s :: !l
    )
    nametable;
  let l' = Sort.list ( <= ) !l in
    List.fold_right 
      (fun x ll -> if ll <> [] & List.hd ll = x then ll else x::ll)
      l' 
      []
;;




(**********************************************************************)
(* ANALYZING DEPENDENCIES *********************************************)
(**********************************************************************)




(* File i Phrase ph *)
let enter_direct_dependencies i ph =
  let idents = ph.ident_set in
  List.iter
    (fun id ->
       try
         let nd = Hashtbl.find nametable id in
	 let deps = ( !files ).( i ).direct_deps in
           if (not (List.mem nd.file deps)) &
	      (i <> nd.file or nd.line > ph.pline) 
           then
             ( !files ).( i ).direct_deps <- nd.file :: deps
       with
         Not_found -> ()
    )
    idents
;;


let get_direct_dependencies i =
  ( !files ).( i ).direct_deps
;;


let get_indirect_dependencies i =
  let rec find_deps (dont:int list) (todo:int list) : int list =
  ( if todo = [] then [] else
    let (next : int list) = 
      List.flatten (List.map (fun j -> ( !files ).( j ).direct_deps) todo) in
    let (next_nodups : int list) =
      List.fold_left (fun l x -> if List.mem x l or List.mem x dont then l 
                                 else x::l) 
                     [] next in
    let (rec_idir_deps : int list) = 
      find_deps (todo @ next_nodups) next_nodups in
    let (idir_deps_nodups : int list) =
      List.fold_left (fun l x -> if List.mem x l then l else x::l) 
                     [] (next_nodups @ rec_idir_deps) in
    idir_deps_nodups
  ) in
  find_deps [] [i]
;;

		         



(**********************************************************************)
(* ANALYZING A FILE PHRASE BY PHRASE **********************************)
(**********************************************************************)

(* Global variables, especially command line parameters and related 
 * variables
 *)

let arg_definitions = ref false;; (* command line switches *)
let arg_crossref = ref false;;
let arg_comments = ref false;;
let arg_dependencies = ref false;;
let arg_types = ref false;;
let arg_html = ref false;;

let nuprl_pipe_dir = ref "/usr/local/lib/nuprl";;
  (* may be customized *)

let html_symbols = ref "gif";;
  (* may be customized *)

let html_symbols_from = ref 128;;
let html_symbols_to = ref 207;;

let nuprl_input_pipe = (ref None : out_channel option ref);;
let nuprl_output_pipe = (ref None : in_channel option ref);;


(* text formatting, including HTML transformation *)


let print_string_indented k s =        (* rather slow *)
  let l = String.length s in
  let p = String.make k ' ' in
  print_string p;
  for i = 0 to l-1 do
    let c = s.[i] in
    if c = '\n' then
      print_string ("\n" ^ p)
    else
      print_char c
  done
;;


let print_indented s i =
  (* don't ask me why I programmed this twice... *)
  (* this function is slightly different than the function above *)
  let inext = ref false in
  for j=0 to String.length s - 1 do
    if s.[j] = '\n' then
    ( inext := true;
      print_char '\n'
    )
    else
    ( if !inext then
        print_string (String.make i ' ');
      print_char s.[j];
      inext := false
    )
  done
;;


let string_indented s i =
  let t = ref "" in
  let inext = ref false in
  for j=0 to String.length s - 1 do
    if s.[j] = '\n' then
    ( inext := true;
      t := !t ^ (String.make 1 '\n')
    )
    else
    ( if !inext then
       t := !t ^  (String.make i ' ');
      t := !t ^ String.make 1 s.[j];
      inext := false
    )
  done;
  !t
;;


let format_html s =
  (* translates the string s which is assumed to use the NuPRL font
   * into an HTML expression.
   * Multiple spaces are not treated specially so the browser displays
   * them as single space.
   *)
  let t = ref "" in
  for i = 0 to String.length s - 1 do
    let c = s.[i] in
    let k = Char.code c in
    if (k>=32 & k <= 127) or c='\n' or c='\t' then
    ( (* there is a fixed translation *)
      match c with
      | '\n'-> t := !t ^ "<br>"
      | '\t'-> t := !t ^ "&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;"
      | ' ' -> t := !t ^ "&#160;"
      | '<' -> t := !t ^ "&lt;"
      | '>' -> t := !t ^ "&gt;"
      | '&' -> t := !t ^ "&amp;"
      | '"' -> t := !t ^ "&quot;"
      | _   -> t := !t ^ (String.make 1 c)
    )
    else
    ( (* translation depends on character set *)
      if k >= !html_symbols_from & k <= !html_symbols_to then
      ( (* a translation is defined *)
	t := !t ^ "<img src=\"" ^ !html_symbols ^ "/char" ^  
             (string_of_int k) ^ ".gif\">"
      )
      else
      ( (* unknown symbol *)
        t := !t ^ "<b>&lt;" ^ (string_of_int k) ^ "&gt;</b>"
      )
    )
  done;
  !t
;;


let test_html_translation () =
  let t = ref "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN//3.2\"><html><body>" in
  for i = 0 to 15 do
    for j = 0 to 15 do
      let s = String.make 1 (Char.chr (i*16+j)) in
      let s_html = format_html s in
      t := !t ^ s_html
    done;
    t := !t ^ "<br>\n"
  done;
  t := !t ^ "</body></html>\n";
  print_string !t
;;


let uppercase_string s =
  let t = String.create (String.length s) in
  for i = 0 to String.length s - 1 do
    let c = s.[i] in
    if (c>='a' & c <='z') then
      t.[i] <- Char.chr (Char.code c - (Char.code 'a' - Char.code 'A'))
    else
      t.[i] <- c
  done;
  t
;;


(* formatting for both html mode and text mode *)


let format_headline s = 
  if !arg_html then
    "<h2>" ^ format_html s ^ "</h2>\n"
  else
    uppercase_string s
;;


let format_newline () =
  if !arg_html then
    "<br>\n"
  else
    "\n"
;;


let format_verbatim s =
  if !arg_html then
    format_html s
  else
    s
;;


let format_list_beginning () =
  if !arg_html then
    "<ul>\n"
  else
    ""
;;


let format_list_end () = 
  if !arg_html then
    "</ul>\n"
  else
    ""
;;


let format_list_item s = 
  if !arg_html then
    "<li>" ^ s ^ "\n"
  else
    "- " ^ string_indented s 2 ^ "\n"
;;


let format_simple_code s =
  if !arg_html then
    "<code>" ^ s ^ "</code>"
  else
    string_indented s 4
;;


let format_keyword_code s =
  if !arg_html then
    "<code><b>" ^ s ^ "</b></code>"
  else
    s
;;


let format_comment s = 
  if !arg_html then
    "<i>" ^ s ^ "</i>"
  else
    string_indented s 4
;;


let format_doc_beginning () =
  if !arg_html then
    "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN//3.2\"><html><body>\n"
  else
    ""
;;


let format_doc_end () =
  if !arg_html then
    "</body></html>\n"
  else
    ""
;;

(* --- *)

let write_names () =
  let l = sorted_list () in
  print_string (format_headline "\nList of names (alphabetically):\n");
  print_string (format_list_beginning());
  List.iter
    ( fun s ->
        let ndl = Hashtbl.find_all nametable s in
        let t = ref "" in
        t := format_simple_code (format_verbatim (s ^ ": "));
        List.iter
          ( fun nd ->
              ( match nd.kind with
                  Odef -> ()
                | Tdef -> t := !t ^ (format_simple_code (format_verbatim "type "))
              );
              
              t := !t ^ (format_verbatim ( !files.(nd.file) ).fname);
              t := !t ^ (format_verbatim ( "/"  ^ string_of_int (nd.line) ^ " "))
          )
          ndl;
        print_string (format_list_item !t)
    )
    l;
  print_string (format_list_end())
;;

(* --- *)

let get_references idlist =
  let il' = List.map (fun id -> try [id, Hashtbl.find nametable id]
			        with
				  Not_found -> [])
                     idlist in
  List.flatten il'
;;


let get_until_end nop =
(* prerr_string("receiving: "); *)
  Unix.clear_nonblock (Unix.descr_of_in_channel nop);
  let s = ref "" in
  let t = " " in
  while input nop t 0 1 = 0 do None done;
  let c = ref (t.[0]) in
  while !c <> end_of_transmission do
    s := !s ^ String.make 1 !c;
(* prerr_char !c; *)
(* flush stderr; *)
    while input nop t 0 1 = 0 do None done;
    c := t.[0]
  done;
(* prerr_newline(); *)
  !s
;;


let get_and_print_nuprl_type name =
  (* ask NuPRL: *)
  if !arg_types then
  ( let nip = match !nuprl_input_pipe with
                None   -> failwith "No connection to NuPRL"
              | Some p -> p in
    let nop = match !nuprl_output_pipe with
                None   -> failwith "No connection to NuPRL"
              | Some p -> p in
(* prerr_string ("sending: " ^ name); *)
(* prerr_newline(); *)
    output_string nip (name ^ String.make 1 end_of_transmission);
(* prerr_string("flushing"); *)
(* prerr_newline(); *)
    flush nip;
    try      
      format_simple_code (format_verbatim (get_until_end nop))
    with
      End_of_file -> format_comment 
                       (format_verbatim  "(connection to NuPRL lost)")
  )
  else ""
;;


let analyze_file i =
  let fname = ( !files ).( i ).fname in
  let ph =    ( !files ).( i ).ph in
  if !arg_definitions then
    print_string 
      (format_headline 
        ("Analysis of file " ^ fname ^ " phrase-by-phrase:\n"));
  if !arg_definitions then
    print_string (format_list_beginning());
  List.iter
    (fun p -> 
       if !arg_definitions then
       ( let item = ref "" in
         if p.type_defs <> [] then
         ( List.iter (fun (n,s) -> item := !item ^ 
                                           (format_keyword_code
                                              (format_verbatim ("type " ^ s)));
                                   if List.mem s !ignored_idents then
                                     item := !item ^ 
                                             (format_comment
                                               (format_verbatim " (IGNORED)"));
                                   item := !item ^ (format_newline())) 
                     p.type_defs
         );
         if p.obj_defs <> [] then
         ( List.iter (fun (n,s) -> item := !item ^ 
                                           (format_keyword_code
                                              (format_verbatim s));
                                   if List.mem s !ignored_idents then
                                     item := !item ^ 
                                             (format_comment
                                               (format_verbatim " (IGNORED)"))
				   else
				     item := !item ^ 
                                             (get_and_print_nuprl_type s);
                                   item := !item ^ (format_newline()))
                     p.obj_defs
         );
         if p.type_defs = [] & p.obj_defs = [] then
           item := !item ^ (format_simple_code (format_verbatim "(nothing)"))
                         ^ (format_newline());
         item := !item ^ 
                 (format_comment
                   (format_verbatim
                     ("\ndefined in line " ^ string_of_int p.pline ^ "\n")));
	 if !arg_comments then
           List.iter
             (fun c ->
                item := !item ^ 
                        (format_comment 
                          (format_verbatim ("\n" ^ c)))
	     )
	     p.comments;
         let rf = get_references p.ident_set in
           if rf <> [] then
           ( item := !item ^ 
                     (format_comment
                       (format_verbatim "\nfound known identifiers:"));
             List.iter
                (fun (id,nd) -> item := !item ^ 
                                        (format_simple_code
                                          (format_verbatim
                                            ( ( if nd.kind = Tdef then
                                                  "\ntype "
                                                 else
                                                  "\n"
                                              ) ^ 
                                              id
					    )
                                          )
                                        ) ^ 
                                          (format_verbatim
                                            (" from file " ^ 
                                               ( !files ).(nd.file).fname ^
                                               " line " ^ string_of_int nd.line
                                            )
                                          )
                )
                rf
           );
           print_string (format_list_item (!item ^ format_newline()))
       );
       enter_direct_dependencies i p;
       enter_phrase i p
    )
    ph;
  if !arg_definitions then
    print_string (format_list_end());
  if !arg_dependencies then
  ( print_string 
      (format_headline ("\nDirect dependencies of file " ^ fname ^ ":\n"));
    print_string (format_list_beginning());
    List.iter
      (fun j ->
         print_string (format_list_item (( ! files ).( j ).fname ^ "\n"))
      )
      (get_direct_dependencies i);
    print_string (format_list_end());
    print_string (format_newline());
    print_string 
      (format_headline ("Indirect dependencies of file " ^ fname ^ ":\n"));
    print_string (format_list_beginning());
    List.iter
      (fun j ->
         print_string (format_list_item (( ! files ).( j ).fname ^ "\n"))
      )
      (get_indirect_dependencies i);
    print_string (format_list_end())
  );
  print_string (format_newline())
;;


(**********************************************************************)
(* ANALYSIS OF ALL GIVEN FILES IN TURN ********************************)
(**********************************************************************)


let analyze_all_files fnames =
  if !arg_types then
  ( prerr_string ("Opening pipe " ^ !nuprl_pipe_dir ^ 
                  "/nuprlinput for writing...");
    prerr_newline();
    let nip = open_out_gen  [Open_wronly;
		             Open_binary;
		             Open_nonblock] 
			    0 
			    (!nuprl_pipe_dir ^ "/nuprlinput") in
    nuprl_input_pipe := Some nip;
    prerr_string ("Opening pipe " ^ !nuprl_pipe_dir ^ 
                  "/nuprloutput for reading...");
    prerr_newline();
    let nop = open_in_gen  [Open_rdonly;
                            Open_binary;
		            Open_nonblock]
			   0
			   (!nuprl_pipe_dir ^ "/nuprloutput") in
    nuprl_output_pipe := Some nop;
    prerr_string "Testing connection to NuPRL...";
    prerr_newline();
    output_string nip ("0" ^ String.make 1 end_of_transmission);
    flush nip;
    try      
      let result = get_until_end nop in
      if result = "0 : int\n" then
      ( prerr_string "Connection works.";
	prerr_newline()
      )
      else
      ( prerr_string ("Bad answer: " ^ result) ;
        prerr_newline();
        failwith "No connection to NuPRL"
      )
    with
      End_of_file -> prerr_string "Got EOF immediately: bad";
                     prerr_newline();
                     failwith "No connection to NuPRL"
  );


  print_string (format_doc_beginning());
  clear_nametable();
  set_files fnames;
  for i = 0 to (List.length fnames) - 1 do
    analyze_file i
  done;

  ( match !nuprl_input_pipe with
      None   -> ()
    | Some p -> close_out p
  );
  ( match !nuprl_output_pipe with
      None   -> ()
    | Some p -> close_in p
  );

  if !arg_crossref then
    write_names();
  print_string (format_doc_end())
;;


(**********************************************************************)
(* INTERPRET THE COMMAND LINE *****************************************)
(**********************************************************************)


(* Parsen der Kommandozeile *)

(* TODO: -ignore "..." --> Symbol ignorieren *)

let parse_cmd_line () =
  let fnames = ref [] in
  Arg.parse [ "-definitions", Arg.Set arg_definitions, 
                     "    Print global definitions.";
              "-comments",  Arg.Set arg_comments,
                     "    Print comments for definitions.";
              "-dependencies", Arg.Set arg_dependencies,
                     "    Print file dependencies.";
              "-ignore", Arg.String (fun s -> ignored_idents := s :: !ignored_idents),
                     "ident    Ignore identifier.";
              "-crossref",  Arg.Set arg_crossref,  
                     "    Print cross-reference of all names.";
              "-types", Arg.Set arg_types,
                     "    Enable the use of a NuPRL process as type server and print types";
              "-nuprlpipes", Arg.String (fun s -> nuprl_pipe_dir := s),
                     "dir    Set the directory containing the pipes to communicate with NuPRL";
              "-html", Arg.Set arg_html,
                     "    Format output as HTML";
              "-htmlsymbols", Arg.String (fun s -> html_symbols := s),
                     "    Directory with GIF pictures of the NuPRL font";
              "-htmlsymbolsfrom", Arg.Int (fun i -> html_symbols_from := i),
                     "    Beginning of the range of characters transformed into GIFs";
              "-htmlsymbolsto", Arg.Int (fun i -> html_symbols_from := i),
                     "    End of the range of characters transformed into GIFs"
            ]
            (fun s ->  fnames := !fnames @ [s])
            "Analyzer for Nuprl ML code (C) 1997 Gerd Stolpmann. List of options:";
  analyze_all_files !fnames
;;


(**********************************************************************)
(* MAIN PROGRAM *******************************************************)
(**********************************************************************)

parse_cmd_line();; 
(* test_html_translation();; *)



(* TODO:
 * -definitions / -defxref trennen
 *)

(* TODO: Ausgabe der Vorkommen von idents, die keine Schluesselworte sind.
 * TODO: Aufbau der Tabellen:
 *  - File -> (name -> zeile der definition)
 *  - name -> (File * zeile der definition) [ggf. list]
 *  - name -> (File * zeile des Vorkommens) list
 *  - File -> (File/Zeile der Definitions jedes vorkommenden Namens)
 * 
 * Ausgabe:
 *  I. File by File, Phrase by Phrase:
 *    File: abc
 *    Definierte Objekte: (alphabetisch sortiert)
 *      - abc1 in Zeile ... benutzt zur Definition ...
 *      - abc2 in Zeile ... benutzt zur Definition ...
 *    Definierte Typen: (alphabetisch sortiert)
 *      - t1 in Zeile ...
 *      - t2 in Zeile ...
 * 
 * II. name by name (alphabetisch sortiert)
 *    Name1: Objektdef in File/Zeile, Typdef in File/Zeile, Vorkommen ...
 *    Name2: ...
 *)

(**********************************************************************)

