
{ (* header *) 
  open Mllexeraux
}

rule get_token =
  parse [ 'A'-'Z' 'a'-'z' ] [ '_' 'A'-'Z' 'a'-'z' '0'-'9' '\'' ]* 
        { (* ident *)
          let s = Lexing.lexeme lexbuf in
            Ident s
        }

      | '$' [ 'A'-'Z' 'a'-'z' ] [ '_' 'A'-'Z' 'a'-'z' '0'-'9' '\'' ]* 
	{ (* dollared ident *)
	  let s = Lexing.lexeme lexbuf in
	    Ident (String.sub s 1 (String.length s - 1))
        }

      | [ '0'-'9' ] [ '0'-'9' ]*
        { (* decimal integer *) 
          let s = Lexing.lexeme lexbuf in
          Integer_constant (int_of_string s)
        }

      | '`' ( [^ '`' '\'' '\\'] | ( '\\' _ ) )+ '`'
        { (* tok constants *) 
          let s = Lexing.lexeme lexbuf in
          let t = (String.sub s 1 (String.length s - 2)) in
	  let n = count_lines s in
            Lines (n, Tok_constant (remove_backslashes_in_tok t))
        }

      | '`' ( [^ '`' '\'' '\\'] | ( '\\' _ ) )* '\''
        { (* var constants *)
          let s = Lexing.lexeme lexbuf in
          let t = (String.sub s 1 (String.length s - 2)) in
	  let n = count_lines s in
            Lines (n, Var_constant (remove_backslashes_in_tok t))
        }

      | "``" ( [^ '`' '\\'] | ( '\\' _ ) )* "``"
        { (* tok list constants *) 
          let s = Lexing.lexeme lexbuf in
          let t = (String.sub s 2 (String.length s - 4)) in
          let n = count_lines s in 
            Lines (n, Toklist_constant (split_list (remove_backslashes_in_tok t)))
        }

      | '"' ( [^ '"' '\\'] | ( '\\' _ ) )* '"'
        { (* string *) 
          let s = Lexing.lexeme lexbuf in
 	  let t = (String.sub s 1 (String.length s - 2)) in
          let n = count_lines s in
            Lines (n, String_constant (remove_backslashes_in_string t))
        }

      | [ ' ' '\t' '\r' ]
        { (* Leerzeichen *) 
          Nulltoken
        }

      | '\n'
        { (* naechste Zeile --> zaehlen! *)
          Lines (1, Nulltoken)
        }

      | "%" [^ '%']* "%"
        { (* comment *)
          let s = Lexing.lexeme lexbuf in
          let n = count_lines s in
            Lines (n, Comment s)
        }

      | ( [^ '%' '0'-'9' ' ' '\n' '\t' '\r' '"' '`' 'A'-'Z' 'a'-'z' '$'] |
          ";;" )
        { (* Symbol *) 
          let s = Lexing.lexeme lexbuf in
          Symbol s
        }
  
      | eof
        { Endofstream }

{ (* trailer *) }
