;;; This code reads ML commands from a named pipe and returns
;;; the results to another named pipe after evaluation.

(setq *input-pipe-name* "/usr/local/lib/nuprl/nuprlinput")
(setq *output-pipe-name* "/usr/local/lib/nuprl/nuprloutput")
(setq *end-of-transmission* (code-char 255))

;;; NOTE: This code was developed with CLISP. The CLISP implementation
;;; notes say that opening special files does not use buffers. This
;;; is very important for a proper function.

(defun start-service ()
  (let ((eos (cons nil nil))             ; marks end of stream
        (input-pipe (open *input-pipe-name* :direction :input )) ; may block
	(output-pipe (open *output-pipe-name* :direction :output))
	(line ""))                       ; the current line
    ; while the input pipe is not at its end...
    (do ((c (read-char input-pipe nil eos)
	     (read-char input-pipe nil eos)))
	 ((eq c eos))
;	 (princ (char-code c) *standard-output*)
;	 (finish-output *standard-output*)
	 (if (eq c *end-of-transmission*) 
	     ( ; then:
	       ; got an EOT character. Evalutate the line and return
	       ; the result.
	      let ((result 
		    (handle-process-err
		     #'(lambda (err-str) 
			 (concatenate 'string "ERROR: " err-str))
                     (ML (itext-term line)))))
	       (princ (concatenate 'string "got: " line) *standard-output*)
	       (terpri *standard-output*)
	       (princ result output-pipe)
	       (princ *end-of-transmission* output-pipe)
	       (finish-output output-pipe)
	       (setq line ""))
	   (; else
	     setq line (concatenate 'string line 
				    (make-string 1 :initial-element c)))))
    (close input-pipe)
    (close output-pipe))
  )
